import pytest

from terka.domain.user import User
from terka.domain.task import Task
from terka.domain.project import Project


@pytest.fixture
def user():
    return User("am")


@pytest.fixture
def task():
    return Task("test_task")


def test_user_init(user):
    assert user.name == "am"


def test_user_create_project(user):
    project = user.create_project("New Project")
    assert project.name == "New Project"


@pytest.mark.skip("Failing")
def test_user_add_task_to_project(user, task):
    project = user.assign_task_to_project("New Project", task)
    assert len(project.tasks) == 1
    assert project.tasks[0].name == "test_task"


def test_user_set_task_due_date(user, task):
    due_date = "2022-01-01"
    user.set_task_due_date(task, due_date)
    assert task.due_date == due_date


def test_user_set_task_incorrect_due_date(user, task):
    due_date = ""
    with pytest.raises(ValueError):
        user.set_task_due_date(task, due_date)


def test_user_can_create_task(user):
    task = user.create_task("new_task")
    assert task.name == "new_task"
    assert task.created_by == "am"
    assert task.project is None


def test_user_can_create_task_and_assign_project(user):
    task = user.create_task(name="new_task", project_id = 2)
    assert task.project == 2




from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import (
    Table,
    MetaData,
    Column,
    Integer,
    String,
    Date,
    DateTime,
    Boolean,
    Enum,
    ForeignKey,
    event,
)
from sqlalchemy.orm import mapper, registry, relationship, validates, declarative_base, backref

from terka.domain.task import Task, TaskStatus, TaskPriority
from terka.domain.user import User
from terka.domain.project import Project, ProjectStatus
from terka.domain.event_history import TaskEvent, ProjectEvent, EventType
from terka.domain.commentary import TaskCommentary, ProjectCommentary, EpicCommentary, StoryCommentary, SprintCommentary
from terka.domain.notes import TaskNote, ProjectNote, EpicNote, StoryNote, SprintNote
from terka.domain.tag import BaseTag, TaskTag, ProjectTag
from terka.domain.collaborators import TaskCollaborator, ProjectCollaborator
from terka.domain.sprint import Sprint, SprintStatus, SprintTask
from terka.domain.time_tracker import TimeTrackerEntry
from terka.domain.epic import Epic, EpicTask, EpicStatus
from terka.domain.story import Story, StoryTask, StoryStatus

from terka.domain.external_connectors.asana import AsanaTask, AsanaProject

Base = declarative_base()
metadata = MetaData()

tasks = Table("tasks", metadata,
              Column("id", Integer, primary_key=True, autoincrement=True),
              Column("name", String(255)),
              Column("creation_date", DateTime, nullable=True),
              Column("modification_date", DateTime, nullable=True),
              Column("description", String(1000), nullable=True),
              Column("project", ForeignKey("projects.id"), nullable=True),
              Column("assignee", ForeignKey("users.id"), nullable=True),
              Column("created_by", ForeignKey("users.id"), nullable=True),
              Column("due_date", Date, nullable=True),
              Column("status", Enum(TaskStatus)),
              Column("priority", Enum(TaskPriority)))

projects = Table("projects", metadata,
                 Column("id", Integer, primary_key=True, autoincrement=True),
                 Column("name", String(255)), Column("description",
                                                     String(255)),
                 Column("status", Enum(ProjectStatus)))

task_events = Table(
    "task_events",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("task", ForeignKey("tasks.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("type", Enum(EventType)),
    Column("old_value", String(225)),
    Column("new_value", String(225)),
)

project_events = Table(
    "project_events",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("project", ForeignKey("projects.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("type", Enum(EventType)),
    Column("old_value", String(225)),
    Column("new_value", String(225)),
)

commentaries = Table(
    "commentaries",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("source", String(50)),
    Column("task", ForeignKey("tasks.id"), nullable=True),
    Column("element_id", Integer),
    Column("date", DateTime, nullable=False),
    Column("text", String(225)),
)

task_commentaries = Table(
    "task_commentaries",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("task", ForeignKey("tasks.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("text", String(225)),
)

project_commentaries = Table(
    "project_commentaries",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("project", ForeignKey("projects.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("text", String(225)),
)

epic_commentaries = Table(
    "epic_commentaries",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("epic", ForeignKey("epics.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("text", String(225)),
)

story_commentaries = Table(
    "story_commentaries",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("story", ForeignKey("stories.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("text", String(225)),
)

sprint_commentaries = Table(
    "sprint_commentaries",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("sprint", ForeignKey("sprints.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("text", String(225)),
)

task_notes = Table(
    "task_notes",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("task", ForeignKey("tasks.id"), nullable=True),
    Column("created_by", ForeignKey("users.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("name", String(225)),
    Column("text", String(1000)),
)

project_notes = Table(
    "project_notes",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("project", ForeignKey("projects.id"), nullable=True),
    Column("created_by", ForeignKey("users.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("name", String(225)),
    Column("text", String(1000)),
)

epic_notes = Table(
    "epic_notes",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("epic", ForeignKey("epics.id"), nullable=True),
    Column("created_by", ForeignKey("users.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("name", String(225)),
    Column("text", String(1000)),
)

story_notes = Table(
    "story_notes",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("story", ForeignKey("stories.id"), nullable=True),
    Column("created_by", ForeignKey("users.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("name", String(225)),
    Column("text", String(1000)),
)

sprint_notes = Table(
    "sprint_notes",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("sprint", ForeignKey("sprints.id"), nullable=True),
    Column("created_by", ForeignKey("users.id"), nullable=True),
    Column("date", DateTime, nullable=False),
    Column("name", String(225)),
    Column("text", String(1000)),
)

users = Table("users", metadata,
              Column("id", Integer, primary_key=True, autoincrement=True),
              Column("name", String(50)))

tags = Table("tags", metadata,
             Column("id", Integer, primary_key=True, autoincrement=True),
             Column("text", String(50)))

task_tags = Table("task_tags", metadata,
                  Column("id", Integer, primary_key=True, autoincrement=True),
                  Column("task", ForeignKey("tasks.id"), nullable=True),
                  Column("tag", ForeignKey("tags.id"), nullable=True))

project_tags = Table(
    "project_tags", metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("project", ForeignKey("projects.id"), nullable=True),
    Column("tag", ForeignKey("tags.id"), nullable=True))

task_collaborators = Table(
    "task_collaborators", metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("task", ForeignKey("tasks.id"), nullable=True),
    Column("collaborator", ForeignKey("users.id"), nullable=True))

project_collaborators = Table(
    "project_collaborators", metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("project", ForeignKey("projects.id"), nullable=True),
    Column("collaborator", ForeignKey("users.id"), nullable=True))

sprints = Table("sprints", metadata,
                Column("id", Integer, primary_key=True, autoincrement=True),
                Column("start_date", Date, nullable=False),
                Column("end_date", Date, nullable=False),
                Column("status", Enum(SprintStatus)),
                Column("goal", String(225), nullable=True))

sprint_tasks = Table(
    "sprint_tasks",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("task", ForeignKey("tasks.id"), nullable=False),
    Column("sprint", ForeignKey("sprints.id"), nullable=False),
    Column("story_points", Integer, nullable=False),
    Column("is_active_link", Boolean, nullable=False),
)

time_tracker_entries = Table(
    "time_tracker_entries",
    metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("creation_date", DateTime, nullable=True),
    Column("task", ForeignKey("tasks.id"), nullable=False),
    Column("time_spent_minutes", Integer, nullable=False),
)

epics = Table("epics", metadata,
              Column("id", Integer, primary_key=True, autoincrement=True),
              Column("name", String(255)),
              Column("creation_date", DateTime, nullable=True),
              Column("description", String(1000), nullable=True),
              Column("project", ForeignKey("projects.id"), nullable=True),
              Column("assignee", ForeignKey("users.id"), nullable=True),
              Column("status", Enum(EpicStatus)),
              Column("created_by", ForeignKey("users.id"), nullable=True))

epic_tasks = Table("epic_tasks", metadata,
                   Column("id", Integer, primary_key=True, autoincrement=True),
                   Column("task", ForeignKey("tasks.id"), nullable=False),
                   Column("epic", ForeignKey("epics.id"), nullable=False))

stories = Table("stories", metadata,
                Column("id", Integer, primary_key=True, autoincrement=True),
                Column("name", String(255)),
                Column("creation_date", DateTime, nullable=True),
                Column("description", String(1000), nullable=True),
                Column("project", ForeignKey("projects.id"), nullable=True),
                Column("assignee", ForeignKey("users.id"), nullable=True),
                Column("status", Enum(StoryStatus)),
                Column("created_by", ForeignKey("users.id"), nullable=True))

story_tasks = Table(
    "story_tasks", metadata,
    Column("id", Integer, primary_key=True, autoincrement=True),
    Column("task", ForeignKey("tasks.id"), nullable=False),
    Column("story", ForeignKey("stories.id"), nullable=False))
# class SprintTasks(Base):
#     __tablename__ = "sprint_tasks"
#     id = Integer(primary_key=True, autoincrement=True)
#     task_id = Integer(nullable=False)
#     sprint_id = Integer(nullable=False)
#     ta

asana_tasks = Table(
    "external_connectors.asana.tasks",
    metadata,
    # Column("id", Integer, primary_key=True, autoincrement=True),
    Column("project", ForeignKey("projects.id")),
    Column("id", ForeignKey("tasks.id"), nullable=False, primary_key=True),
    Column("asana_task_id", String(20)),
    Column("sync_date", DateTime, nullable=True))

asana_projects = Table(
    "external_connectors.asana.projects",
    metadata,
    # Column("id", Integer, primary_key=True, autoincrement=True),
    Column("id", ForeignKey("projects.id"), nullable=False, primary_key=True),
    Column("asana_project_id", String(20)),
    Column("sync_date", DateTime, nullable=True))


def start_mappers():
    asana_task_mapper = mapper(AsanaTask, asana_tasks)
    asana_project_mapper = mapper(AsanaProject, asana_projects)

    task_commentary_mapper = mapper(TaskCommentary, task_commentaries)
    project_commentary_mapper = mapper(ProjectCommentary, project_commentaries)
    epic_commentary_mapper = mapper(EpicCommentary, epic_commentaries)
    story_commentary_mapper = mapper(StoryCommentary, story_commentaries)
    sprint_commentary_mapper = mapper(SprintCommentary, sprint_commentaries)

    task_note_mapper = mapper(TaskNote, task_notes)
    project_note_mapper = mapper(ProjectNote, project_notes)
    epic_note_mapper = mapper(EpicNote, epic_notes)
    story_note_mapper = mapper(StoryNote, story_notes)
    sprint_note_mapper = mapper(SprintNote, sprint_notes)

    task_event_mapper = mapper(TaskEvent, task_events)
    project_event_mapper = mapper(ProjectEvent, project_events)
    user_mapper = mapper(User, users)

    tag_mapper = mapper(BaseTag, tags)
    task_tag_mapper = mapper(TaskTag,
                             task_tags,
                             properties={
                                 "base_tag":
                                 relationship(tag_mapper,
                                              collection_class=list)
                             })
    project_tag_mapper = mapper(ProjectTag,
                                project_tags,
                                properties={
                                    "base_tag":
                                    relationship(tag_mapper,
                                                 collection_class=list)
                                })
    task_collaborator_mapper = mapper(
        TaskCollaborator,
        task_collaborators,
        properties={"users": relationship(user_mapper, collection_class=list)})
    project_collaborator_mapper = mapper(
        ProjectCollaborator,
        project_collaborators,
        properties={"users": relationship(user_mapper, collection_class=list)})
    time_tracker_mapper = mapper(TimeTrackerEntry, time_tracker_entries)
    task_mapper = mapper(Task,
                         tasks,
                         properties={
                             "collaborators":
                             relationship(task_collaborator_mapper,
                                          collection_class=set,
                                          cascade="all, delete-orphan"),
                             "tags":
                             relationship(task_tag_mapper,
                                          collection_class=set,
                                          cascade="all, delete-orphan"),
                             "commentaries":
                             relationship(task_commentary_mapper,
                                          collection_class=list,
                                          cascade="all, delete-orphan"),
                             "notes":
                             relationship(task_note_mapper,
                                          collection_class=set,
                                          cascade="all, delete-orphan"),
                             "history":
                             relationship(task_event_mapper,
                                          collection_class=list,
                                          cascade="all, delete-orphan"),
                             "time_spent":
                             relationship(time_tracker_mapper,
                                          collection_class=list,
                                          cascade="all, delete-orphan"),
                         })
    epic_tasks_mapper = mapper(EpicTask,
                               epic_tasks,
                               properties={
                                   "tasks":
                                   relationship(task_mapper,
                                                backref=backref("epics"),
                                                collection_class=list)
                               })
    epic_mapper = mapper(Epic,
                         epics,
                         properties={
                             "commentaries":
                             relationship(epic_commentary_mapper,
                                          collection_class=list,
                                          cascade="all, delete-orphan"),
                             "notes":
                             relationship(epic_note_mapper,
                                          collection_class=set,
                                          cascade="all, delete-orphan"),
                             "tasks":
                             relationship(epic_tasks_mapper,
                                          backref=backref("epics"),
                                          collection_class=list)
                         })
    story_tasks_mapper = mapper(StoryTask,
                                story_tasks,
                                properties={
                                    "tasks":
                                    relationship(task_mapper,
                                                 backref=backref("stories"),
                                                 collection_class=list)
                                })
    story_mapper = mapper(Story,
                          stories,
                          properties={
                              "commentaries":
                              relationship(story_commentary_mapper,
                                           collection_class=list,
                                           cascade="all, delete-orphan"),
                              "notes":
                              relationship(story_note_mapper,
                                           collection_class=set,
                                           cascade="all, delete-orphan"),
                              "tasks":
                              relationship(story_tasks_mapper,
                                           collection_class=list)
                          })
    project_mapper = mapper(Project,
                            projects,
                            properties={
                                "collaborators":
                                relationship(project_collaborator_mapper,
                                             collection_class=set,
                                             cascade="all, delete-orphan"),
                                "tasks":
                                relationship(task_mapper,
                                             collection_class=set),
                                "epics":
                                relationship(epic_mapper,
                                             collection_class=set),
                                "stories":
                                relationship(story_mapper,
                                             collection_class=set),
                                "tags":
                                relationship(project_tag_mapper,
                                             collection_class=set,
                                             cascade="all, delete-orphan"),
                                "commentaries":
                                relationship(project_commentary_mapper,
                                             collection_class=list,
                                             cascade="all, delete-orphan"),
                                "notes":
                                relationship(project_note_mapper,
                                             collection_class=set,
                                             cascade="all, delete-orphan"),
                                "history":
                                relationship(project_event_mapper,
                                             collection_class=list)
                            })
    sprint_tasks_mapper = mapper(SprintTask,
                                 sprint_tasks,
                                 properties={
                                     "tasks":
                                     relationship(task_mapper,
                                                  backref=backref("sprints"),
                                                  collection_class=list)
                                 })
    sprint_mapper = mapper(Sprint,
                           sprints,
                           properties={
                               "commentaries":
                               relationship(sprint_commentary_mapper,
                                            collection_class=list,
                                            cascade="all, delete-orphan"),
                               "notes":
                               relationship(sprint_note_mapper,
                                            collection_class=set,
                                            cascade="all, delete-orphan"),
                               "tasks":
                               relationship(sprint_tasks_mapper,
                                            collection_class=list),
                           })



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, clear_mappers

from terka.adapters.orm import metadata, start_mappers

@pytest.fixture
def in_memory_db():
    engine = create_engine("sqlite:///:memory:")
    metadata.create_all(engine)
    return engine

@pytest.fixture
def session(in_memory_db):
    start_mappers()
    yield sessionmaker(bind=in_memory_db)()
    clear_mappers()



from typing import Dict, List, Optional
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from sqlalchemy import text

from terka.domain.task import Task


def _n_days_ago(n: int) -> str:
    return (datetime.today() - timedelta(n)).strftime("%Y-%m-%d")


def time_spent(
    session,
    tasks: List[Task],
    start_date: str = _n_days_ago(7),
    end_date: str = _n_days_ago(0)
) -> List[Dict[str, str]]:
    tasks = ",".join([str(task.id) for task in tasks])
    results = session.execute(
        text(f"""
    SELECT
        STRFTIME("%Y-%m-%d", creation_date) AS date,
        SUM(time_spent_minutes) AS time_spent_hours
    FROM time_tracker_entries
    WHERE
        task IN ({tasks})
        AND creation_date >= "{start_date}"
        AND creation_date <= "{end_date}"
    GROUP BY 1
    """))
    return [dict(r) for r in results]


def sprint_task_ids(session,
                    sprint_id: Optional[int] = None) -> List[Dict[int, int]]:
    results = session.execute(
        """
    SELECT
        task, sprint
    FROM sprint_tasks
    WHERE
        sprint = :sprint_id
    """, dict(sprint_id=sprint_id))
    return [dict(r) for r in results]


def external_connectors_asana_projects(session) -> List[Dict[int, str]]:
    results = session.execute("""
    SELECT
        id, asana_project_id, sync_date
    FROM 'external_connectors.asana.projects'
    """)
    return [dict(r) for r in results]


def external_connectors_asana_tasks(session,
                                    project_id: str) -> List[Dict[int, str]]:
    results = session.execute(
        """
    SELECT
        id, asana_task_id, sync_date
    FROM 'external_connectors.asana.tasks'
    WHERE project = :project_id
    """, dict(project_id=project_id))
    return [dict(r) for r in results]


def status_changes(session, sprint_id: int):
    query = f"""
        SELECT DISTINCT T.id AS task,
        T.status AS current_status,
        STRFTIME("%Y-%m-%d", E.date) AS date,
        FIRST_VALUE(E.old_value) OVER (PARTITION BY T.id, STRFTIME("%Y-%m-%d", E.date) ORDER by E.date RANGE BETWEEN UNBOUNDED PRECEDING AND
                UNBOUNDED FOLLOWING) AS first_status_for_date,
        LAST_VALUE(E.new_value) OVER (PARTITION BY T.id, STRFTIME("%Y-%m-%d", E.date) ORDER by E.date RANGE BETWEEN UNBOUNDED PRECEDING AND
                UNBOUNDED FOLLOWING) AS last_status_for_date
        FROM tasks AS T
        LEFT JOIN (SELECT * FROM task_events WHERE type = "STATUS") AS E
            ON T.id = E.task
        WHERE T.id IN (SELECT task FROM sprint_tasks WHERE sprint = {sprint_id})

        """
    return pd.read_sql_query(query, session.get_bind()).replace({np.nan: None})



from datetime import datetime


class BaseCommentary:

    def __init__(self, text: str, date: datetime = datetime.now(), **kwargs: str) -> None:
        self.text = text
        self.date = date

    def __repr__(self):
        return f"<Commentary> {self.text}"


class TaskCommentary(BaseCommentary):

    def __init__(self,
                 id: int,
                 text: str,
                 date: datetime = datetime.now(), **kwargs: str) -> None:
        self.task = id
        super().__init__(text=text,
                         date=date)


class ProjectCommentary(BaseCommentary):

    def __init__(self, id: int,
                 text: str,
                 date: datetime = datetime.now(), **kwargs: str) -> None:
        self.project = id
        super().__init__(text=text,
                         date=date)


class EpicCommentary(BaseCommentary):

    def __init__(self, id: int,
                 text: str,
                 date: datetime = datetime.now(), **kwargs: str) -> None:
        self.epic = id
        super().__init__(text=text,
                         date=date)


class StoryCommentary(BaseCommentary):

    def __init__(self, id: int,
                 text: str,
                 date: datetime = datetime.now(), **kwargs: str) -> None:
        self.story = id
        super().__init__(text=text,
                         date=date)


class SprintCommentary(BaseCommentary):

    def __init__(self, id: int,
                 text: str,
                 date: datetime = datetime.now(), **kwargs: str) -> None:
        self.sprint = id
        super().__init__(text=text,
                         date=date)



import pytest
from datetime import datetime

from src.domain.task import Task, TaskStatus
from src.domain.user import User
from src.domain.event_history import Event
from src.domain.commentary import Commentary


def test_user_can_load(session):
    session.execute("INSERT INTO users (name) VALUES "
                    '("user_1")')
    expected = [User("user_1")]
    print(session.query(User).all()[0].name)
    assert session.query(User).all() == expected


def test_can_save_tasks(session):
    new_task = Task("task_1")
    session.add(new_task)
    session.commit()
    rows = list(session.execute('SELECT id, name, status FROM "tasks"'))
    assert rows == [(1, "task_1", "BACKLOG")]


def test_can_save_tasks_with_project(session):
    new_task = Task(name="task_1", project="new project")
    session.add(new_task)
    session.commit()
    rows = list(session.execute('SELECT id, name, project, status FROM "tasks"'))
    assert rows == [(1, "task_1", "new project", "BACKLOG")]


def test_can_save_users(session):
    new_user = User("user_2")
    session.add(new_user)
    session.commit()

    rows = list(session.execute('SELECT id, name FROM "users"'))
    assert rows == [(1, "user_2")]


def test_can_save_one_event(session):
    date = datetime.now()
    event = Event("tasks", 1, date, "status", "BACKLOG", "TODO")
    session.add(event)
    session.commit()
    rows = list(
        session.execute(
            'SELECT id, source, element_id, date, type, old_value, new_value FROM "events"'
        ))
    assert rows == [(1, "tasks", 1, date.strftime("%Y-%m-%d %H:%M:%S.%f"),
                     "status", "BACKLOG", "TODO")]


def test_can_save_multiple_events(session):
    date1 = datetime.now()
    event1 = Event("tasks", 1, date1, "status", "BACKLOG", "TODO")
    date2 = datetime.now()
    event2 = Event("tasks", 1, date2, "status", "TODO", "DONE")
    session.add(event1)
    session.add(event2)
    session.commit()
    rows = list(session.execute('SELECT id FROM "events"'))
    assert rows == [(1, ), (2, )]


def test_can_save_one_commentary(session):
    date = datetime.now()
    commentary = Commentary(1, "Test commentary", date)
    session.add(commentary)
    session.commit()
    rows = list(
        session.execute('SELECT task_id, date, text FROM "commentaries"'))
    assert rows == [(1, date.strftime("%Y-%m-%d %H:%M:%S.%f"),
                     "Test commentary")]


def test_can_save_multiple_commentaries(session):
    date1 = datetime.now()
    commentary1 = Commentary(1, "Test commentary", date1 )
    date2 = datetime.now()
    commentary2 = Commentary(1, "Test commentary", date2)
    session.add(commentary1)
    session.add(commentary2)
    session.commit()
    rows = list(session.execute('SELECT id FROM "commentaries"'))
    assert rows == [(1, ), (2, )]



import pytest

from datetime import datetime
from terka.domain.task import Task, TaskStatus


@pytest.fixture
def task():
    return Task(name="name")


def test_create_task_with_name_and_creation_date():
    now = datetime.now()
    task = Task(name = "name", creation_date=now)
    assert task.name == "name"
    assert task.description is None
    assert task.creation_date == now
    assert task.project == None
    assert task.status == "BACKLOG"
    assert task.assignee == None



def test_create_task_with_description():
    task2 = Task(name="name", description="description")
    assert task2.description == "description"


def test_fails_empty_name():
    with pytest.raises(ValueError):
        task = Task("")


def test_fails_incorrect_date():
    with pytest.raises(ValueError):
        task = Task("test", creation_date="2020-01-01")



from typing import Dict, Optional
from dataclasses import dataclass


@dataclass
class Command:
    entity: str
    payload: Dict[str, str]
    # def __init__(self, entity: str, payload: Optional[Dict[str, str]]):
    #     self.entity = entity
    #     self.payload = payload

    # def __repr__(self) -> str:
    #     return f"{self.__class__.__name__}(entity={self.entity}, payload={self.payload})"


@dataclass
class Edit(Command):
    ...


@dataclass
class Create(Command):
    ...


@dataclass
class Update(Command):
    ...


@dataclass
class Complete(Command):
    ...


@dataclass
class Start(Command):
    ...


@dataclass
class Create(Command):
    ...


@dataclass
class List(Command):
    ...


@dataclass
class Show(Command):
    ...


@dataclass
class Get(Command):
    ...


@dataclass
class Collaborate(Command):
    ...


@dataclass
class Track(Command):
    ...


@dataclass
class Tag(Command):
    ...


@dataclass
class Add(Command):
    ...


@dataclass
class Note(Command):
    ...


@dataclass
class Comment(Command):
    ...


@dataclass
class Delete(Command):
    ...


@dataclass
class Connect(Command):
    ...


@dataclass
class Report(Command):
    ...



import argparse
from datetime import datetime, date
from io import StringIO
import re
import os
from pathlib import Path
from operator import attrgetter
import sys
import yaml
import logging
from pprint import pprint

import rich
from rich.console import Console
from rich.table import Table
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, clear_mappers

from terka.adapters.orm import metadata, start_mappers
from terka.adapters.repository import SqlAlchemyRepository

from terka.domain import commands
from terka.domain.commands import CommandHandler
from terka.utils import (
    format_task_dict,
    process_command,
    update_task_dict,
    create_task_dict,
)
from terka.service_layer import services


def init_db(home_dir):
    engine = create_engine(f"sqlite:////{home_dir}/.terka/tasks.db")
    metadata.create_all(engine)
    return engine


class TerkaInitError(Exception):
    ...


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("command", nargs="?")
    parser.add_argument("entity", nargs="?")
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("-v", "--version", dest="version", action="store_true")
    args = parser.parse_known_args()
    args, kwargs = args
    command, entity = args.command, args.entity
    if args.version:
        import pkg_resources

        version = pkg_resources.require("terka")[0].version
        print(f"terka version {version}")
        exit()
    if args.command == "config":
        if "--show" in kwargs:
            pprint(services.get_config())
        else:
            services.update_config(create_task_dict(kwargs))
        exit()

    logger = logging.getLogger(__name__)
    console = Console()

    home_dir = os.path.expanduser("~")
    config = load_config(home_dir)
    task_id = config.get("task_id")
    project_name = config.get("project_name")
    if task_id or project_name:
        focus_type = "task" if task_id else "project"
        logger.warning("Using terka in focus mode")
        logger.warning(f"Current focus is {focus_type}: {task_id or project_name}")

    task_dict = format_task_dict(config, args, kwargs)
    logger.debug(task_dict)

    service_command_handler = services.ServiceCommandHander(home_dir, config, console)
    service_command_handler.execute(command, entity, task_dict)

    engine = init_db(home_dir)
    start_mappers()
    Session = sessionmaker(engine)

    with Session() as session:
        file_handler = logging.FileHandler(filename=f"{home_dir}/.terka/terka.log")
        file_handler.setLevel(logging.DEBUG)
        stdout_handler = logging.StreamHandler(stream=sys.stdout)
        stdout_handler.setLevel(logging.WARNING)
        handlers = [file_handler, stdout_handler]
        logging.basicConfig(
            format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
            handlers=handlers,
            level=args.loglevel.upper(),
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        repo = SqlAlchemyRepository(session)
        command_handler = CommandHandler(repo)

        task_dict = format_task_dict(config, args, kwargs)
        task_dict = update_task_dict(task_dict, repo)
        logger.debug(task_dict)

        if file_path := task_dict.get("file"):
            with open(file_path, "r") as f:
                lines = [line.rstrip() for line in f if line.rstrip()]
                for line in lines:
                    entry = line.strip().split("::")
                    if not entry:
                        continue
                    match entry:
                        case [project, name, description]:
                            task_dict = {
                                "name": name,
                                "project": project,
                                "description": description,
                            }
                        case [project, name]:
                            task_dict = {"name": name, "project": project}
                        case [name]:
                            task_dict = {"name": name}
                        case _:
                            raise Exception("Unknown format")
                    task_dict = update_task_dict(task_dict, repo)
                    if entity.startswith("s/"):
                        entity = "stories"
                    elif entity.startswith("e/"):
                        entity = "epics"
                    command_handler.execute(command, entity, task_dict)
            exit()
        is_interactive = False
        if not command and not entity:
            is_interactive = True
            interactive_message = (
                f"[green]>>> Running terka in an interactive mode[/green]"
            )
            if task_id:
                interactive_message = f"{interactive_message} (focus task {task_id})"
            elif project_name:
                interactive_message = (
                    f"{interactive_message} (focus project {project_name})"
                )
            console.print(interactive_message)
            command = input("enter command: ")
            command, entity, task_dict = process_command(command, config, repo)
        try:
            command_handler.execute(command, entity, task_dict)
        except (ValueError, commands.TerkaException) as e:
            console.print(f"[red]{e}[/red]")

        while is_interactive:
            interactive_message = (
                f"[green]>>> Running terka in an interactive mode[/green]"
            )
            config = load_config(home_dir)
            task_id = config.get("task_id")
            project_name = config.get("project_name")
            if task_id:
                interactive_message = f"{interactive_message} (focus task {task_id})"
            elif project_name:
                interactive_message = (
                    f"{interactive_message} (focus project {project_name})"
                )
            console.print(interactive_message)
            command = input("enter command: ")
            command, entity, task_dict = process_command(command, config, repo)
            if command[0] == "q":
                exit()
            try:
                command_handler.execute(command, entity, task_dict)
            except commands.TerkaException as e:
                console.print(f"[red]{e}[/red]")
                exit()


def load_config(home_dir):
    try:
        with open(f"{home_dir}/.terka/config.yaml", encoding="utf-8") as f:
            config = yaml.safe_load(f)
        return config
    except FileNotFoundError:
        raise TerkaInitError("call `terka init` to initialize terka")


if __name__ == "__main__":
    main()



from enum import Enum


class ProjectStatus(Enum):
    DELETED = 0
    ACTIVE = 1
    ON_HOLD = 2
    COMPLETED = 3


class Project:

    _next_project_id = 1

    def __init__(self,
                 name: str,
                 description: str = None,
                 created_by: str = None,
                 status: str = "ACTIVE"):
        self._project_id = Project._next_project_id
        Project._next_project_id += 1
        self.name = name
        self.description = description
        self.created_by = created_by
        self.status = status

    def _validate_status(self, status):
        if status not in [
                s.name for s in ProjectStatus if s.name != "DELETED"
        ]:
            raise ValueError(f"{status} is invalid status")
        else:
            return status

    def __str__(self):
        return f"<Project {self.id}>: {self.name} {self.tasks}"



from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from .task import Task


class StoryStatus(Enum):
    ACTIVE = 1
    COMPLETED = 2


class Story:

    def __init__(self,
                 name: str,
                 description: str = None,
                 creation_date: datetime = datetime.now(),
                 project: int = None,
                 assignee: int = None,
                 status: str = "ACTIVE",
                 created_by: int = None,
                 **kwargs) -> None:
        if not name:
            raise ValueError("task name cannot be empty!")
        if not isinstance(creation_date, datetime):
            raise ValueError(
                "creation_date should be of type datetime.datetime!")
        self.name = name
        self.creation_date = creation_date
        self.description = description
        self.project = project
        self.assignee = assignee
        self.created_by = created_by
        self.status = status
        self.is_completed = False

    def complete(self, tasks) -> None:
        incompleted_tasks = list()
        for task in tasks:
            if task.tasks.status.name != "DONE":
                incompleted_tasks.append(task)
        if incompleted_tasks:
            logging.warning("[story %d]: %d tasks haven't been completed",
                            self.id, len(incompleted_tasks))
        self.is_completed = True


@dataclass
class StoryTask:
    story: int
    task: int



from enum import Enum
from datetime import datetime


class TaskStatus(Enum):
    DELETED = 0
    BACKLOG = 1
    TODO = 2
    IN_PROGRESS = 3
    REVIEW = 4
    DONE = 5


class TaskPriority(Enum):
    UNKNOWN = 0
    LOW = 1
    NORMAL = 2
    HIGH = 3
    URGENT = 4


class Task:

    _next_task_id = 1

    def __init__(self,
                 name: str,
                 description: str = None,
                 creation_date: datetime = datetime.now(),
                 project: int = None,
                 assignee: int = None,
                 created_by: int = None,
                 due_date: datetime = None,
                 status: str = "BACKLOG",
                 priority: str = "NORMAL",
                 **kwargs):
        if not name:
            raise ValueError("task name cannot be empty!")
        if not isinstance(creation_date, datetime):
            raise ValueError(
                "creation_date should be of type datetime.datetime!")
        self._task_id = Task._next_task_id
        Task._next_task_id += 1

        self.name = name
        self.creation_date = creation_date
        self.description = description
        self.project = project
        self.assignee = assignee
        self.created_by = created_by
        if due_date and due_date.date() < datetime.today().date():
            raise ValueError(
                f"Due date {due_date.date()} cannot be less than today")
        else:
            self.due_date = due_date
        self.status = self._validate_status(status)
        if priority and priority not in [p.name for p in TaskPriority]:
            raise ValueError(f"{priority} is invalid priority")
        else:
            self.priority = priority

    def _validate_status(self, status):
        if status and status not in [
                s.name for s in TaskStatus if s.name != "DELETED"
        ]:
            raise ValueError(f"{status} is invalid status")
        else:
            return status

    @property
    def total_time_spent(self):
        if self.time_spent:
            return sum([t.time_spent_minutes for t in self.time_spent])
        return 0

    def __repr__(self):
        return f"<Task {self.id}>: {self.name}, {self.status.name}, {self.creation_date}"



from typing import Any, Dict, Optional
import logging
import rich
from rich.console import Console
from rich.table import Table
import os
import re
import yaml

from terka.adapters.repository import AbsRepository
from terka.domain.project import Project
from terka.domain.user import User
import terka.domain._commands as commands


def update_config(update_pair: Dict[str, Any]):
    config = get_config()
    config.update(update_pair)
    with open(f"{home_dir}/.terka/config.yaml", "w") as f:
        yaml.dump(config, f)


def get_config():
    home_dir = os.path.expanduser('~')
    with open(f"{home_dir}/.terka/config.yaml", "r", encoding="utf-8") as f:
        config = yaml.safe_load(f)
    return config


def lookup_project_id(project_name: str, repo: AbsRepository) -> int:
    projects = project_name.split(",")
    if project_name.startswith("NOT"):
        projects = [
            f"NOT:{project}"
            for project in project_name.replace("NOT:", "").split(",")
        ]
        negate = True
    else:
        projects = project_name.split(",")
        negate = False
    returned_projects = []
    returned_projects_set = set()
    for project_name in projects:
        if not (project := repo.list(Project, project_name)):
            print(
                f"Creating new project: {project_name}. Do you want to continue (Y/n)?"
            )
            answer = input()
            while answer.lower() != "y":
                print("Provide a project name: ")
                project_name = input()
                print(
                    f"Creating new project: {project_name}. Do you want to continue (Y/n)?"
                )
                answer = input()
            project = Project(project_name)
            repo.add(project)
            repo.session.commit()
        if isinstance(project, list):
            project_ids = [project.id for project in project]
            if negate:
                if returned_projects_set:
                    returned_projects_set = returned_projects_set.intersection(
                        set(project_ids))
                else:
                    returned_projects_set = returned_projects_set.union(
                        set(project_ids))

            else:
                returned_projects.extend(project_ids)
        else:
            returned_projects.append(project.id)
    if returned_projects_set:
        return list(returned_projects_set)
    if len(returned_projects) == 1:
        return returned_projects[0]
    return returned_projects


def lookup_user_id(user_name: str, repo: AbsRepository) -> Optional[int]:
    user = repo.list(User, {"name": user_name})
    if user:
        return user[0].id
    return None


def lookup_user_name(user_id: str, repo: AbsRepository) -> Optional[str]:
    user = repo.list(User, {"id": user_id})
    if user:
        return user[0].name
    return None


def lookup_project_name(project_id: int, repo: AbsRepository) -> Optional[str]:
    project = repo.list(Project, {"id": project_id})
    if project:
        return project[0]
    return None



class CommandHander:

    def __init__(self, home_dir, config, console=Console()):
        self.home_dir = home_dir
        self.config = config
        self.console =  console

    def execute(self, command: commands.Command, entity, task_dict):
        # TODO: create registry of commands
        handler = commands_registry[command]
        handler.handle(entity, task_dict)
        


class ServiceCommandHander:

    def __init__(self, home_dir, config, console=Console()):
        self.home_dir = home_dir
        self.config = config
        self.console =  console


    def execute(self, command, entity, task_dict):
        is_service_command = True
        if command == "init":
            self.init_terka()
        elif command == "focus":
            self.focus(entity, task_dict)
        elif command == "unfocus":
            self.unfocus()
        elif command == "log":
            self.log(task_dict)
        else:
            is_service_command = False
        if is_service_command:
            exit()


    def init_terka(self,
                   terka_folder: str = ".terka",
                   default_user: str = "admin",
                   default_config: str = "config.yaml") -> None:
        path = os.path.join(self.home_dir, terka_folder)
        if not os.path.exists(path):
            answer = input(
                f"Do you want to init terka in this directory {path}? [Y/n]")
            if "y" in answer.lower():
                Path(path).mkdir(parents=True, exist_ok=True)
                with open(os.path.join(path, "config.yaml"), "w") as f:
                    yaml.dump({"user": default_user}, f)
            elif "n" in answer.lower():
                path = input("Specify full path to the terka directory: ")
                os.mkdirs(path)
            else:
                exit()
        elif not os.path.exists(os.path.join(path, default_config)):
            answer = input(f"Config.yaml not found in {path}, Create it? [Y/n]")
            if "y" in answer.lower():
                with open(os.path.join(path, default_config), "w") as f:
                    yaml.dump({"user": default_user}, f)
            else:
                exit()
        else:
            print("Terka directory already exist.")

    def focus(self, entity_type, kwargs):
        if entity_type == "tasks":
            self.config["task_id"] = kwargs["id"]
            if "project_name" in self.config.keys():
                del self.config["project_name"]
        if entity_type == "projects":
            self.config["project_name"] = kwargs["id"]
            if "task_id" in self.config.keys():
                del self.config["task_id"]
        with open(f"{self.home_dir}/.terka/config.yaml",
                  "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        logging.info("<focus> %s: %s", entity_type, "")

    def unfocus(self):
        if "task_id" in self.config.keys():
            del self.config["task_id"]
        if "project_name" in self.config.keys():
            del self.config["project_name"]
        with open(f"{self.home_dir}/.terka/config.yaml",
                  "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        loging.info("<unfocus> %s: %s", "", "")

    def log(self, kwargs):
        table = Table(box=rich.box.SIMPLE)
        with open(f"{self.home_dir}/.terka/terka.log", "r") as f:
            head = f.readlines()
        for column in ("date", "source", "level", "message"):
            table.add_column(column)
        num_log_entries = int(kwargs.get("num_log_entries", 10))
        tail = head[-num_log_entries:]
        for row in tail[::-1]:
            info, message = row.split("] ")
            date, source, level = info.split("][")
            table.add_row(re.sub("\[", "", date), source, level, message)
        self.console.print(table)



from datetime import datetime
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Header, Static, Input


# TODO: extract out of printer
def format_time_spent(time_spent: int) -> str:
    time_spent_hours = time_spent // 60
    time_spent_minutes = time_spent % 60
    if time_spent_hours and time_spent_minutes:
        time_spent = f"{time_spent_hours}H:{time_spent_minutes}M"
    elif time_spent_hours:
        time_spent = f"{time_spent_hours}H:00M"
    elif time_spent_minutes:
        time_spent = f"00H:{time_spent_minutes}M"
    else:
        time_spent = ""
    return time_spent



class Comment(Widget):
    value = reactive("text")

    def render(self) -> str:
        return f"Comment: {self.value}"


class TerkaTask(App):
    BINDINGS = [("q", "quit", "Quit")]
    CSS_PATH = "vertical_layout.css"

    def __init__(self,
                 entity,
                 project,
                 is_completed: bool = False,
                 history=None,
                 commentaries=None):
        super().__init__()
        self.entity = entity
        self.project = project
        self.is_completed = is_completed
        self.history = history
        self.commentaries = commentaries
        self.is_overdue = datetime.today().date(
        ) > entity.due_date if entity.due_date else False

    def compose(self) -> ComposeResult:
        yield Header()
        task_text = f"[bold]Task {self.entity.id}: {self.entity.name}[/bold]"
        if self.is_completed:
            yield Static(task_text, classes="header_completed", id="header")
        elif self.is_overdue:
            yield Static(task_text, classes="header_overdue", id="header")
        else:
            yield Static(task_text, classes="header_simple", id="header")
        yield Static(f"Project: [bold]{self.project}[/bold]", classes="transp")
        yield Static(f"Status: [bold]{self.entity.status.name}[/bold]",
                     classes="transp")
        yield Static(f"Priority: [bold]{self.entity.priority.name}[/bold]",
                     classes="transp")
        created_days_ago = (datetime.now() - self.entity.creation_date).days
        creation_message = f"{created_days_ago} days ago" if created_days_ago > 1 else "today"
        yield Static(f"Created {creation_message}", classes="header_simple")
        if self.is_completed:
            completion_date = self._get_completion_date()
            if completion_date:
                completed_days_ago = (datetime.now() - completion_date).days
                completion_message = f"{completed_days_ago} days ago" if completed_days_ago > 1 else "today"
                yield Static(
                    f"Completed {completion_message} ({completion_date.strftime('%Y-%m-%d')})",
                    classes="header_simple")
            else:
                yield Static("Completion date unknown",
                             classes="header_simple")
        else:
            yield Static(f"Due date: {self.entity.due_date}",
                         classes="header_simple")
        if sprints := self.entity.sprints:
            sprint_id = ",".join(str(s.sprint) for s in sprints)
            story_points = sprints[-1].story_points
            yield Static(
                f"Sprints: [bold]{sprint_id}[/bold], "
                f"SP: [bold]{story_points}[/bold], "
                f"T: [bold]{format_time_spent(self.entity.total_time_spent)}[/bold]",
                classes="transp")
        else:
            yield Static(f"Not in sprint", classes="transp")
        if epics := self.entity.epics:
            epic_ids = ",".join([str(e.epic) for e in epics])
            yield Static(f"Epics: [bold]{epic_ids}[/bold]", classes="transp")
        else:
            yield Static(f"Not in epic", classes="transp")
        if stories := self.entity.stories:
            story_ids = ",".join([str(s.story) for s in stories])
            yield Static(f"Story: [bold]{story_ids}[/bold]", classes="transp")
        else:
            yield Static(f"Not in story", classes="transp")
        if tags := self.entity.tags:
            tags = ",".join(
                t.base_tag.text for t in tags
                if not t.base_tag.text.startswith(("epic", "story", "sprint")))
            yield Static(f"Tags: [bold]{tags}[/bold]", classes="transp")
        else:
            yield Static(f"No tags", classes="transp")
        if collaborators := self.entity.collaborators:
            collaborators = ",".join(c.users.name for c in collaborators)
            yield Static(f"Collaborators: [bold]{collaborators}[/bold]",
                         classes="transp")
        else:
            yield Static(f"No collaborators", classes="transp")
        description_message = self.entity.description or ""
        if description_message:
            yield Static(
                f"[italic]Description...\n[/italic]{description_message}",
                classes="body",
                id="desc")
        else:
            yield Static("[italic]No description...[/italic]",
                         classes="body",
                         id="desc")
        if self.commentaries:
            comms = [
                f"[italic]{comment.date.date()}[/italic]: {comment.text}"
                for comment in self.commentaries
            ]
            comms_message = "\n".join(comms)
            yield Static(f"[italic]Comments...\n\n[/italic]{comms_message}",
                         classes="body",
                         id="history")
        else:
            yield Static("", classes="body", id="history")
        yield Input(placeholder="Add a comment", classes="body", id="comment")
        # yield Comment()

    # def on_input_changed(self, event: Input.Changed) -> None:
    #     self.query_one(Comment).text = event.value

    def _get_completion_date(self):
        for event in self.history:
            if event.new_value == "DONE":
                return event.date
